"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
clipping functions
"""

import logging
import numpy as np
import pathlib
import shapely
import subprocess
from cartopy.io import shapereader

from osm_flex.config import POLY_DIR, OSMCONVERT_PATH
LOGGER = logging.getLogger(__name__)


def get_admin1_shapes(country):
    """Provide Natural Earth registry info and shape files for countries

    Parameters
    ----------
    country : str
        string of ISO 3166 code

    Returns
    -------
    country_shapes : dict
        Shapes (according to Natural Earth) of admin1 regions of country
        with name as keys
    """

    if not isinstance(country, str):
        LOGGER.error("country needs to be of type str")
        raise TypeError("Invalid type for input parameter 'country'")
    admin1_file = shapereader.natural_earth(resolution='10m',
                                            category='cultural',
                                            name='admin_1_states_provinces')
    admin1_recs = shapereader.Reader(admin1_file)
    admin1_shapes = {}
    for rec in admin1_recs.records():
        if rec.attributes['adm0_a3'] == country:
            name = rec.attributes['name_en']
            admin1_shapes[name] = rec.geometry
    if not admin1_shapes:
        raise LookupError(f'natural_earth records are empty for country {country}')
    return admin1_shapes

def get_country_shape(country):
    """Provide Natural Earth registry info and shape files for admin1 regions
    of chosen country

    Parameters
    ----------
    country : str
        string of ISO 3166 code

    Returns
    -------
    country_shape : (multi-)polygon
        Shape of the country according to Natural Earth.
    """

    if not isinstance(country, str):
        LOGGER.error("country needs to be of type str")
        raise TypeError("Invalid type for input parameter 'country'")
    admin0_file = shapereader.natural_earth(resolution='10m',
                                            category='cultural',
                                            name='admin_0_countries')
    admin0_recs = shapereader.Reader(admin0_file)
    for rec in admin0_recs.records():
        if rec.attributes['ADM0_A3'] == country:
            return rec.geometry
    raise LookupError(f'natural_earth records are empty for country {country}')


def _simplify_shapelist(geom_list, thres=None):
    """
    remove tiny shapes and simplify outlines to save on file size for
    .poly files
    """
    if thres is None:
        thresh = 0.1 if shapely.unary_union(geom_list).area > 1 else 0.01
    geom_list = [geom for geom in geom_list if geom.area>thresh]
    return [geom.simplify(tolerance=0.01, preserve_topology=True) for
            geom in geom_list]


def _shapely2poly(geom_list, filename):
    """
    Convert list of shapely (multi)polygon(s) into .poly files needed for
    osmosis to generate cut-outs from bigger osm.pbf files.

    Parameters
    ---------
    geom_list : list
        list of polygon, polygons or multipolygons containing a (complex) shape
        to be cut out of a bigger file
    filename : pathlib.Path or str
        output filename including directory path.

    Returns
    -------
    None

    Note
    ----
    For more info on what .poly files are (incl. several tools for
    creating them), see
    https://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Format
    """
    filename = pathlib.Path(filename).with_suffix('.poly')
    if filename.exists():
        raise ValueError(f'File {filename} already exists, aborting.')

    # start writing the .poly file
    file = open(filename, 'w')
    file.write('Polygons' + "\n")

    # loop over the different polygons, get their exterior and write the
    # coordinates of the ring to the .poly file
    i = 0
    for shape in geom_list:
        if shape.geom_type == 'MultiPolygon':
            polygons = shape.geoms
        elif shape.geom_type == 'Polygon':
            polygons = [shape]

        for polygon in polygons:
            polygon = np.array(polygon.exterior.coords)
            file.write(str(i) + "\n")
            i += 1
            for ring in polygon:
                file.write("    " + str(ring[0]) + "     " + str(ring[1]) +"\n")

            # close the ring of one subpolygon if done
            file.write("END" +"\n")

    # close the file when done
    file.write("END" +"\n")
    file.close()

def _build_osmconvert_cmd(shape, osmpbf_clip_from, osmpbf_output):
    """
    builds osmconvert command for clipping

    Parameters
    -----------
    shape : list or str or pathlib.Path
        list containing [xmin, ymin, xmax, ymax] for a bounding box or
        a string/Path to the .poly file path delimiting the bounds.
    osmpbf_clip_from: str or pathlib.Path
        file path to planet.osm.pbf or other osm.pbf file to clip
    osmpbf_output : str or pathlib.Path
        file path (incl. name & ending) under which extract will be stored
    """

    if isinstance(shape, (pathlib.Path, str)):
        return [str(OSMCONVERT_PATH), str(osmpbf_clip_from), f'-B={str(shape)}',
                '--complete-ways', '--complete-multipolygons',
                f'-o={osmpbf_output}'
                ]
    if isinstance(shape[0], (float, int)):
        return [str(OSMCONVERT_PATH), str(osmpbf_clip_from),
                '-b='+str(shape[0])+','+str(shape[1])+','+str(shape[2])+','+str(shape[3]),
                '--complete-ways', '--complete-multipolygons',
                f'-o={osmpbf_output}'
                ]

def _osmconvert_clip(shape, osmpbf_clip_from, osmpbf_output,
                     overwrite=False):

    osmpbf_clip_from = pathlib.Path(osmpbf_clip_from)
    if not osmpbf_clip_from.suffix:
        osmpbf_clip_from = osmpbf_clip_from.with_suffix('.osm.pbf')
    if not osmpbf_clip_from.is_file():
        raise ValueError(f"OSM file {osmpbf_clip_from} to clip from not found.")

    if ((not pathlib.Path(osmpbf_output).is_file()) or
        (pathlib.Path(osmpbf_output).is_file() and overwrite)):

        LOGGER.info("""File doesn`t yet exist or overwriting old one.
                    Assembling osmosis command.""")
        cmd = _build_osmconvert_cmd(shape, osmpbf_clip_from, osmpbf_output)

    LOGGER.info('''Extracting from larger file...
                This will take a while''')

    return subprocess.run(cmd, stdout=subprocess.PIPE,
                          universal_newlines=True)


def _build_osmosis_cmd(shape, osmpbf_clip_from, osmpbf_output):
    """
    builds osmosis command for clipping

    Parameters
    -----------
    shape : list or str or pathlib.Path
        list containing [xmin, ymin, xmax, ymax] for a bounding box or
        a string/Path to the .poly file path delimiting the bounds.
    osmpbf_clip_from: str or pathlib.Path
        file path to planet.osm.pbf or other osm.pbf file to clip
    osmpbf_output : str or pathlib.Path
        file path (incl. name & ending) under which extract will be stored
    """
    if isinstance(shape, (pathlib.Path, str)):
        return ['osmosis', '--read-pbf', 'file='+str(osmpbf_clip_from),
            '--bounding-polygon', 'file='+str(shape), '--write-pbf',
            'file='+str(osmpbf_output)]
    if isinstance(shape[0], (float, int)):
        return['osmosis', '--read-pbf', 'file='+str(osmpbf_clip_from),
            '--bounding-box', f'top={shape[3]}', f'left={shape[0]}',
            f'bottom={shape[1]}', f'right={shape[2]}',
            '--write-pbf', 'file='+str(osmpbf_output)]

    raise ValueError('''shape does not have the correct format.
                        Only bounding boxes, shapely (multi-)polygons or
                        filepaths to .poly files are allowed''')


def _osmosis_clip(shape, osmpbf_clip_from, osmpbf_output,
                  overwrite=False):
    """
    Runs the command line tool osmosis to cut out all map info within
    shape (bounding box or poygon(s)), from a bigger parent file.

    If your device doesn't have osmosis yet, see installation instructions:
    https://wiki.openstreetmap.org/wiki/Osmosis/Installation

    Parameters
    -----------
    shape : list or str or pathlib.Path
        list containing [xmin, ymin, xmax, ymax] for a bounding box or
        a string/Path to the .poly file path delimiting the bounds.
    osmpbf_clip_from: str or pathlib.Path
        file path to planet.osm.pbf or other osm.pbf file to clip
    osmpbf_output : str or pathlib.Path
        file path (incl. name & ending) under which extract will be stored
    overwrite : bool
        default is False. Whether to overwrite files if they already exist.

    Returns
    -------
    None or subprocess
    """
    osmpbf_clip_from = pathlib.Path(osmpbf_clip_from)
    if not osmpbf_clip_from.suffix:
        osmpbf_clip_from = osmpbf_clip_from.with_suffix('.osm.pbf')
    if not osmpbf_clip_from.is_file():
        raise ValueError(f"OSM file {osmpbf_clip_from} to clip from not found.")

    if ((not pathlib.Path(osmpbf_output).is_file()) or
        (pathlib.Path(osmpbf_output).is_file() and overwrite)):

        LOGGER.info("""File doesn`t yet exist or overwriting old one.
                    Assembling osmosis command.""")

        cmd = _build_osmosis_cmd(shape, osmpbf_clip_from, osmpbf_output)

        LOGGER.info('''Extracting from larger file...
                    This will take a while''')

        return subprocess.run(cmd, stdout=subprocess.PIPE,
                            universal_newlines=True)

    raise ValueError(f"File {osmpbf_output} already exists. Abort.")


def clip_from_bbox(bbox, osmpbf_clip_from, osmpbf_output,
                   overwrite=False, kernel='osmosis'):
    """
    get OSM raw data from abounding-box, which is extracted
    from a bigger (e.g. the planet) file.

    Parameters
    ----------
    bbox : list
        bounding box [xmin, ymin, xmax, ymax]
    osmpbf_clip_from: str or pathlib.Path
        file path to planet.osm.pbf or other osm.pbf file to clip
    osmpbf_output : str or pathlib.Path
        file path (incl. name & ending) under which extract will be stored
    overwrite : bool
        default is False. Whether to overwrite files if they already exist.
    kernel : str
        name of the clipping kernel: 'osmconvert' or 'osmosis'
        Default is 'osmosis'.
    Note
    ----
    This function uses the command line tool osmosis to cut out new
    osm.pbf files from the original ones.
    Installation instructions (windows, linux, apple) - see
    https://wiki.openstreetmap.org/wiki/Osmosis/Installation
    """
    # TODO: allow for osmpbf_output to be only file name & save in default DIR
    if kernel == 'osmosis':
        _osmosis_clip(bbox, osmpbf_clip_from, osmpbf_output, overwrite)
        return
    if kernel == 'osmconvert':
        _osmconvert_clip(bbox, osmpbf_clip_from, osmpbf_output, overwrite)
        return
    raise ValueError(f"Kernel '{kernel}' is not valid. Abort.")


def clip_from_poly(poly_file, osmpbf_clip_from, osmpbf_output,
                   overwrite=False, kernel='osmosis'):
    """
    get OSM raw data from a custom shape defined in .poly file which is clipped
    from a .osm.pbf file.

    Parameters
    ----------
    poly_file : str
        file path to a .poly file
    osmpbf_output : str or pathlib.Path
        file path (incl. name & ending) under which extract will be stored
    clip_from_file : str or pathlib.Path
        file path to planet-latest.osm.pbf. Will download & store it as
        indicated, if doesn`t yet exist.
        Default is DATA_DIR/planet-latest.osm.pbf
    overwrite : bool
        default is False. Whether to overwrite files if they already exist.
    kernel : str
        name of the clipping kernel: 'osmconvert' or 'osmosis'
        Default is 'osmosis'.

    Note
    ----
    This function uses by default the command line tool osmosis
    to cut out new osm.pbf files from the original ones.
    Installation instructions (windows, linux, macos) - see
    https://wiki.openstreetmap.org/wiki/Osmosis/Installation
    """
    if kernel == 'osmosis':
        _osmosis_clip(poly_file, osmpbf_clip_from, osmpbf_output, overwrite)
        return
    if kernel == 'osmconvert':
        _osmconvert_clip(poly_file, osmpbf_clip_from, osmpbf_output, overwrite)
        return
    raise ValueError(f"Kernel '{kernel}' is not valid. Abort.")


def clip_from_shapes(shape_list, osmpbf_clip_from, osmpbf_output,
                     overwrite=False, kernel='osmosis'):
    """
    get OSM raw data from a custom shape defined by a list of polygons
    which is extracted from the entire OSM planet file.
    The list of shapes first needs to be converted to a .poly file and then
    passed back to the function (under the hood, a temporary file is created
    and deleted upon completion again).

    Parameters
    ----------
    shape_list : list
        list of (Multi-)Polygon(s) that define the shape which should be cut,
        as e.g. obtained
    osmpbf_output : str
        Full file path under which the clipped data will be stored.
    osmpbf_clip_from : str or pathlib.Path
        file path (including filename) to the *.osm.pbf file to clip from.
    overwrite : bool
        default is False. Whether to overwrite files if they already exist.
    kernel : str
        name of the clipping kernel: 'osmconvert' or 'osmosis'
        Default is 'osmosis'.

    Note
    ----
    This function uses the command line tool osmosis to cut out new
    osm.pbf files from the original ones.
    Installation instructions (windows, linux, apple) - see
    https://wiki.openstreetmap.org/wiki/Osmosis/Installation
    """

    shape_list = _simplify_shapelist(shape_list)

    poly_file = POLY_DIR / 'temp_shp.poly'

    _shapely2poly(shape_list, poly_file)
    if kernel == 'osmosis':
        _osmosis_clip(poly_file, osmpbf_clip_from, osmpbf_output, overwrite)
        poly_file.unlink()
        return
    if kernel == 'osmconvert':
        _osmconvert_clip(poly_file, osmpbf_clip_from, osmpbf_output, overwrite)
        poly_file.unlink()
        return

    raise ValueError(f"Kernel '{kernel}' is not valid. Abort.")



"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
constants and configurations
"""
from pathlib import Path


# =============================================================================
# FILE AND FOLDER PATHS
# =============================================================================
OSM_CONFIG_FILE = Path(__file__).parent.joinpath('osmconf.ini')
OSM_DIR = Path(Path.home(), 'osm')
OSMCONVERT_PATH = OSM_DIR / 'osmconvert'
OSM_DATA_DIR = OSM_DIR.joinpath("osm_bpf")
POLY_DIR = OSM_DIR.joinpath("poly")
EXTRACT_DIR = OSM_DIR.joinpath("extracts")

# =============================================================================
# URLS
# =============================================================================

GEOFABRIK_URL = 'https://download.geofabrik.de/'
PLANET_URL = 'https://planet.openstreetmap.org/pbf/planet-latest.osm.pbf'


# =============================================================================
# DICTIONARIES
# =============================================================================

"""
dictionary for the generation of the correct download api-address at
geofabrik.de, relating ISO3-country codes to the region & written-out name.
Adapted from the GitHub repo osm_clipper (https://github.com/ElcoK/osm_clipper)
Used by download.get_country_geofabrik().

Note: A few small countries will be downloaded as a multi-country file, as
indicated in the comments.

Note: "special" ISO-3 codes - Canary Islands (IC), Asian part of Russia (RUS-A),
European part of Russia (RUS-E)
"""
DICT_GEOFABRIK = {
   'AFG' : ('asia','afghanistan'),
   'ALB' : ('europe','albania'),
   'DZA' : ('africa','algeria'),
   'AND' : ('europe','andorra'),
   'AGO' : ('africa','angola'),
   'BEN' : ('africa', 'benin'),
   'BWA' : ('africa', 'botswana'),
   'BFA' : ('africa', 'burkina-faso'),
   'BDI' : ('africa', 'burundi'),
   'CMR' : ('africa', 'cameroon'),
   'IC' : ('africa', 'canary-islands'),
   'CPV' : ('africa', 'cape-verde'),
   'CAF' : ('africa', 'central-african-republic'),
   'TCD' : ('africa', 'chad'),
   'COM' : ('africa', 'comores'),
   'COG' : ('africa', 'congo-brazzaville'),
   'COD' : ('africa', 'congo-democratic-republic'),
   'DJI' : ('africa', 'djibouti'),
   'EGY' : ('africa', 'egypt'),
   'GNQ' : ('africa', 'equatorial-guinea'),
   'ERI' : ('africa', 'eritrea'),
   'ETH' : ('africa', 'ethiopia'),
   'GAB' : ('africa', 'gabon'),
   'GMB' : ('africa', 'senegal-and-gambia'),  #TOGETHER WITH SENEGAL
   'GHA' : ('africa', 'ghana'),
   'GIN' : ('africa', 'guinea'),
   'GNB' : ('africa', 'guinea-bissau'),
   'CIV' : ('africa', 'ivory-coast'),
   'KEN' : ('africa', 'kenya'),
   'LSO' : ('africa', 'lesotho'),
   'LBR' : ('africa', 'liberia'),
   'LBY' : ('africa', 'libya'),
   'MDG' : ('africa', 'madagascar'),
   'MWI' : ('africa', 'malawi'),
   'MLI' : ('africa', 'mali'),
   'MRT' : ('africa', 'mauritania'),
   'MAR' : ('africa', 'morocco'),
   'MOZ' : ('africa', 'mozambique'),
   'NAM' : ('africa', 'namibia'),
   'NER' : ('africa', 'niger'),
   'NGA' : ('africa', 'nigeria'),
   'RWA' : ('africa', 'rwanda'),
   'SHN' : ('africa', 'saint-helena-ascension-and-tristan-da-cunha'),
   'STP' : ('africa', 'sao-tome-and-principe'),
   'SEN' : ('africa', 'senegal-and-gambia'),  #TOGETHER WITH THE GAMBIA
   'SYC' : ('africa', 'seychelles'),
   'SLE' : ('africa', 'sierra-leone'),
   'SOM' : ('africa', 'somalia'),
   'ZAF' : ('africa', 'south-africa'),
   'SDN' : ('africa', 'sudan'),
   'SSD' : ('africa', 'south-sudan'),
   'SWZ' : ('africa', 'swaziland'),
   'TZA' : ('africa', 'tanzania'),
   'TGO' : ('africa', 'togo'),
   'TUN' : ('africa', 'tunisia'),
   'UGA' : ('africa', 'uganda'),
   'ZMB' : ('africa', 'zambia'),
   'ZWE' : ('africa', 'zimbabwe'),
   'ARM' : ('asia', 'armenia'),
   'AZE' : ('asia', 'azerbaijan'),
   'BGD' : ('asia', 'bangladesh'),
   'BTN' : ('asia', 'bhutan'),
   'KHM' : ('asia', 'cambodia'),
   'CHN' : ('asia', 'china'),
   'SAU' : ('asia', 'gcc-states'),  #Together with Kuwait, the United Arab Emirates, Qatar, Bahrain, and Oman
   'KWT' : ('asia', 'gcc-states'),  #Together with Saudi Arabia, the United Arab Emirates, Qatar, Bahrain, and Oman
   'ARE' : ('asia', 'gcc-states'),  #Together with Saudi Arabia, Kuwait, Qatar, Bahrain, and Oman
   'QAT' : ('asia', 'gcc-states'),  #Together with Saudi Arabia, Kuwait, the United Arab Emirates, Bahrain, and Oman
   'OMN' : ('asia', 'gcc-states'),  #Together with Saudi Arabia, Kuwait, the United Arab Emirates, Qatar and Oman
   'BHR' : ('asia', 'gcc-states'),  #Together with Saudi Arabia, Kuwait, the United Arab Emirates, Qatar and Bahrain
   'IND' : ('asia', 'india'),
   'IDN' : ('asia', 'indonesia'),
   'IRN' : ('asia', 'iran'),
   'IRQ' : ('asia', 'iraq'),
   'ISR' : ('asia', 'israel-and-palestine'),  # TOGETHER WITH PALESTINE
   'PSE' : ('asia', 'israel-and-palestine'),  # TOGETHER WITH ISRAEL
   'JPN' : ('asia', 'japan'),
   'JOR' : ('asia', 'jordan'),
   'KAZ' : ('asia', 'kazakhstan'),
   'KGZ' : ('asia', 'kyrgyzstan'),
   'LAO' : ('asia', 'laos'),
   'LBN' : ('asia', 'lebanon'),
   'MYS' : ('asia', 'malaysia-singapore-brunei'),  # TOGETHER WITH SINGAPORE AND BRUNEI
   'SGP' : ('asia', 'malaysia-singapore-brunei'),  # TOGETHER WITH MALAYSIA AND BRUNEI
   'BRN' : ('asia', 'malaysia-singapore-brunei'),  # TOGETHER WITH MALAYSIA AND SINGAPORE
   'MDV' : ('asia', 'maldives'),
   'MNG' : ('asia', 'mongolia'),
   'MMR' : ('asia', 'myanmar'),
   'NPL' : ('asia', 'nepal'),
   'PRK' : ('asia', 'north-korea'),
   'PAK' : ('asia', 'pakistan'),
   'PHL' : ('asia', 'philippines'),
   'RUS-A' : ('asia', 'russia'),  # Asian part of Russia
   'KOR' : ('asia', 'south-korea'),
   'LKA' : ('asia', 'sri-lanka'),
   'SYR' : ('asia', 'syria'),
   'TWN' : ('asia', 'taiwan'),
   'TJK' : ('asia', 'tajikistan'),
   'THA' : ('asia', 'thailand'),
   'TKM' : ('asia', 'turkmenistan'),
   'UZB' : ('asia', 'uzbekistan'),
   'VNM' : ('asia', 'vietnam'),
   'YEM' : ('asia', 'yemen'),
   'BHS' : ('central-america', 'bahamas'),
   'BLZ' : ('central-america', 'belize'),
   'CUB' : ('central-america', 'cuba'),
   'GTM' : ('central-america', 'guatemala'),
   'HTI' : ('central-america', 'haiti-and-domrep'),  # TOGETHER WITH DOMINICAN REPUBLIC
   'DOM' : ('central-america', 'haiti-and-domrep'),  # TOGETHER WITH HAITI
   'JAM' : ('central-america', 'jamaica'),
   'HND' : ('central-america', 'honduras'),
   'NIC' : ('central-america', 'nicaragua'),
   'SLV' : ('central-america', 'el-salvador'),
   'CRI' : ('central-america', 'costa-rica'),
   'AUT' : ('europe', 'austria'),
   'BLR' : ('europe', 'belarus'),
   'BEL' : ('europe', 'belgium'),
   'BIH' : ('europe', 'bosnia-herzegovina'),
   'BGR' : ('europe', 'bulgaria'),
   'HRV' : ('europe', 'croatia'),
   'CYP' : ('europe', 'cyprus'),
   'CZE' : ('europe', 'czech-republic'),
   'DNK' : ('europe', 'denmark'),
   'EST' : ('europe', 'estonia'),
   'FRO' : ('europe', 'faroe-islands'),
   'FIN' : ('europe', 'finland'),
   'FRA' : ('europe', 'france'),
   'GEO' : ('europe', 'georgia'),
   'DEU' : ('europe', 'germany'),
   'GBR' : ('europe', 'great-britain'),  # DOES NOT INCLUDE NORTHERN ISLAND
   'GRC' : ('europe', 'greece'),
   'HUN' : ('europe', 'hungary'),
   'ISL' : ('europe', 'iceland'),
   'IRL' : ('europe', 'ireland-and-northern-ireland'),
   'IMN' : ('europe', 'isle-of-man'),
   'ITA' : ('europe', 'italy'),
   'LVA' : ('europe', 'latvia'),
   'LIE' : ('europe', 'liechtenstein'),
   'LTU' : ('europe', 'lithuania'),
   'LUX' : ('europe', 'luxembourg'),
   'MKD' : ('europe', 'macedonia'),
   'MLT' : ('europe', 'malta'),
   'MDA' : ('europe', 'moldova'),
   'MCO' : ('europe', 'monaco'),
   'MNE' : ('europe', 'montenegro'),
   'NLD' : ('europe', 'netherlands'),
   'NOR' : ('europe', 'norway'),
   'POL' : ('europe', 'poland'),
   'PRT' : ('europe', 'portugal'),
   'ROU' : ('europe', 'romania'),
   'RUS-E' : ('europe', 'russia'),  # European part of Russia
   'SRB' : ('europe', 'serbia'),
   'SVK' : ('europe', 'slovakia'),
   'SVN' : ('europe', 'slovenia'),
   'ESP' : ('europe', 'spain'),
   'SWE' : ('europe', 'sweden'),
   'CHE' : ('europe', 'switzerland'),
   'TUR' : ('europe', 'turkey'),
   'UKR' : ('europe', 'ukraine'),
   'CAN' : ('north-america', 'canada'),
   'GRL' : ('north-america', 'greenland'),
   'MEX' : ('north-america', 'mexico'),
   'USA' : ('north-america', 'us'),
   'AUS' : ('australia-oceania', 'australia'),
   'COK' : ('australia-oceania', 'cook-islands'),
   'FJI' : ('australia-oceania', 'fiji'),
   'KIR' : ('australia-oceania', 'kiribati'),
   'MHL' : ('australia-oceania', 'marshall-islands'),
   'FSM' : ('australia-oceania', 'micronesia'),
   'NRU' : ('australia-oceania', 'nauru'),
   'NCL' : ('australia-oceania', 'new-caledonia'),
   'NZL' : ('australia-oceania', 'new-zealand'),
   'NIU' : ('australia-oceania', 'niue'),
   'PLW' : ('australia-oceania', 'palau'),
   'PNG' : ('australia-oceania', 'papua-new-guinea'),
   'PCN' : ('australia-oceania', 'pitcairn-islands'),
   'WSM' : ('australia-oceania', 'samoa'),
   'SLB' : ('australia-oceania', 'solomon-islands'),
   'TON' : ('australia-oceania', 'tonga'),
   'TUV' : ('australia-oceania', 'tuvalu'),
   'VUT' : ('australia-oceania', 'vanuatu'),
   'ARG' : ('south-america', 'argentina'),
   'BOL' : ('south-america', 'bolivia'),
   'BRA' : ('south-america', 'brazil'),
   'CHL' : ('south-america', 'chile'),
   'COL' : ('south-america', 'colombia'),
   'ECU' : ('south-america', 'ecuador'),
   'PRY' : ('south-america', 'paraguay'),
   'PER' : ('south-america', 'peru'),
   'SUR' : ('south-america', 'suriname'),
   'URY' : ('south-america', 'uruguay'),
   'VEN' : ('south-america', 'venezuela'),
}

"""
nested dictionary that contains collections of relevant columns (osm_keys) and
key - value pairs (osm_query) to extract critical infrastructure data from an
osm.pbf file, via the function OSM_FileQuery().retrieve_cis()

Currently implemented for:
    * educational facilities,
    * electric power,
    * food supply,
    * healthcare facilities,
    * natural gas infrastructure,
    * oil infrastructure,
    * road,
    * rail,
    * telecommunications,
    * water supply,
    * wastewater.

Note: If modified, make sure that key exists in osm.config file, under the
respective geometry/-ies.
"""

DICT_CIS_OSM =  {
        'education' : {
            'osm_keys' : ['amenity','building','name'],
            'osm_query' : """building='school' or amenity='school' or
                             building='kindergarten' or 
                             amenity='kindergarten' or
                             building='college' or amenity='college' or
                             building='university' or amenity='university' or
                             building='college' or amenity='college' or
                             building='childcare' or amenity='childcare'"""},
        'healthcare' : {
            'osm_keys' : ['amenity','building','healthcare','name'],
            'osm_query' : """amenity='hospital' or healthcare='hospital' or
                             building='hospital' or building='clinic' or
                             amenity='clinic' or healthcare='clinic' or 
                             amenity='doctors' or healthcare='doctors'
                             """},
        'water' : {
            'osm_keys' : ['man_made','pump','pipeline','emergency','name'],
            'osm_query' : """man_made='water_well' or man_made='water_works' or
                             man_made='water_tower' or
                             man_made='reservoir_covered' or 
                             landuse='reservoir' or
                             (man_made='pipeline' and substance='water') or
                             (pipeline='substation' and substance='water') or
                             pump='powered' or pump='manual' or pump='yes' or
                             emergency='fire_hydrant' or
                             (man_made='storage_tank' and content='water')"""},
        'telecom' : {
            'osm_keys' : ['man_made','tower_type','telecom',
                          'communication_mobile_phone','name'],
            'osm_query' : """tower_type='communication' or man_made='mast' or
                             communication_mobile_phone='*' or
                             telecom='antenna' or
                             telecom='poles' or communication='pole' or
                             telecom='central_office' or 
                             telecom='street_cabinet' or
                             telecom='exchange' or telecom='data_center' or
                             telecom='distribution_point' or 
                             telecom='connection_point' or
                             telecom='line' or communication='line' or
                             utility='telecom'"""},
        'road' :  {
            'osm_keys' : ['highway','name','maxspeed','lanes','surface'],
            'osm_query' : """highway in ('motorway', 'motorway_link', 'trunk', 'trunk_link',
                            'primary', 'primary_link', 'secondary', 'secondary_link', 'tertiary',
                            'tertiary_link', 'residential', 'road', 'unclassified')
                             """},
        'main_road' :  {
            'osm_keys' : ['highway','name','maxspeed','lanes','surface'],
            'osm_query' : """highway in ('primary', 'primary_link', 'secondary',
                             'secondary_link', 'tertiary', 'tertiary_link', 'trunk', 'trunk_link', 
                             'motorway', 'motorway_link')
                            """},
        'rail' : {
            'osm_keys' : ['railway','name','gauge','electrified','voltage'],
            'osm_query' : """railway='rail' or railway='tram' or
                             railway='subway' or railway='narrow_gauge' or
                             railway='light_rail' or
                             railway='station' or railway='platform' or
                             railway='stop' or railway='tram_stop' or
                             railway='signal' or railway='switch'"""},
         'air' : {
             'osm_keys' : ['aeroway','name'],
             'osm_query' : """aeroway='aerodrome'"""},
         'gas' : {
             'osm_keys' : ['man_made','pipeline', 'utility','name'],
             'osm_query' : """(man_made='pipeline' and substance='gas') or
                              (pipeline='substation' and substance='gas') or
                              (man_made='storage_tank' and content='gas') or
                              utility='gas'"""},
        'oil' : {
             'osm_keys' : ['pipeline','man_made','amenity','name'],
             'osm_query' : """(pipeline='substation' and substance='oil') or
                              (man_made='pipeline' and substance='oil') or
                              man_made='petroleum_well' or 
                              man_made='oil_refinery' or
                              amenity='fuel'"""},
        'power' : {
              'osm_keys' : ['power','voltage','utility','name'],
              'osm_query' : """power='line' or power='cable' or
                               power='minor_line' or power='plant' or
                               power='generator' or power='substation' or
                               power='transformer' or
                               power='pole' or power='portal' or 
                               power='tower' or power='terminal' or 
                               power='switch' or power='catenary_mast' or
                               utility='power'"""},
        'wastewater' : {
              'osm_keys' : ['reservoir_type','man_made','utility','natural',
                            'name'],
              'osm_query' : """reservoir_type='sewage' or
                               (man_made='storage_tank' and content='sewage') or
                               (man_made='pipeline' and substance='sewage') or
                               substance='waterwaste' or 
                               substance='wastewater' or
                               (natural='water' and water='wastewater') or
                               man_made='wastewater_plant' or
                               man_made='wastewater_tank' or
                               utility='sewerage'"""},
         'food' : {
             'osm_keys' : ['shop','name'],
             'osm_query' : """shop='supermarket' or shop='greengrocer' or
                              shop='grocery' or shop='general' or 
                              shop='bakery'"""},
                              
        'buildings' : {
            'osm_keys' : ['building','amenity','name'],
            'osm_query' : """building='yes' or building='house' or 
                            building='residential' or building='detached' or 
                            building='hut' or building='industrial' or 
                            building='shed' or building='apartments'"""}
                              }



"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
extraction functions
"""

import logging
import geopandas as gpd
from osgeo import ogr, gdal
import pandas as pd
from pathlib import Path
import shapely
from tqdm import tqdm

from osm_flex.config import DICT_CIS_OSM, OSM_CONFIG_FILE


LOGGER = logging.getLogger(__name__)
DATA_DIR = '' #TODO: dito, where & how to define
gdal.SetConfigOption("OSM_CONFIG_FILE", str(OSM_CONFIG_FILE))


def _query_builder( geo_type, constraint_dict):
    """
    This function builds an SQL query from the values passed to the extract()
    function.

    Parameters
    ---------
    geo_type : str
        Type of geometry to extract. One of [points, lines, multipolygons]
    constraint_dict :  dict

    Returns
    -------
    query : str
        an SQL query string.
    """
    # columns which to report in output
    query =  "SELECT osm_id"
    for key in constraint_dict['osm_keys']:
        query+= ","+ key
    # filter condition(s)
    query+= " FROM " + geo_type + " WHERE " + constraint_dict['osm_query']

    return query

def extract(osm_path, geo_type, osm_keys, osm_query):
    """
    Function to extract geometries and tag info for entires in the OSM file
    matching certain OSM key-value constraints.
    from an OpenStreetMap osm.pbf file.

    Parameters
    ----------
    osm_path : str or Path
        location of osm.pbf file from which to parse
    geo_type : str
        Type of geometry to extract. One of [points, lines, multipolygons]
    osm_keys : list
        a list with all the osm keys that should be reported as columns in
        the output gdf.
    osm_query : str
        query string of the syntax
        "key(='value') (and/or further queries)".
        See examples in DICT_CIS_OSM in case of doubt.

    Returns
    -------
    gpd.GeoDataFrame
        A gdf with all results from the osm.pbf file matching the
        specified constraints.

    Note
    ----
    1) The keys that are searchable are specified in the osmconf.ini file.
    Make sure that they exist in the attributes=... paragraph under the
    respective geometry section.
    For example, to extract multipolygons with building='yes',
    building must be in the attributes under
    the [multipolygons] section of the file. You can find it in the same
    folder as the osm_dataloader.py module is located.
    2) OSM keys that have : in their name must be changed to _ in the
    search dict, but not in the osmconf.ini
    E.g. tower:type is called tower_type, since it would interfere with the
    SQL syntax otherwise, but still tower:type in the osmconf.ini

    See also
    --------
    https://taginfo.openstreetmap.org/ to check what keys and key/value
    pairs are valid.
    https://overpass-turbo.eu/ for a direct visual output of the query,
    and to quickly check the validity. The wizard can help you find the
    correct keys / values you are looking for.
    """
    if not Path(osm_path).is_file():
        raise ValueError(f"the given path is not a file: {osm_path}")
    
    osm_path = str(osm_path)
    constraint_dict = {
        'osm_keys' : osm_keys,
        'osm_query' : osm_query}

    driver = ogr.GetDriverByName('OSM')
    data = driver.Open(osm_path)
    query = _query_builder(geo_type, constraint_dict)
    LOGGER.debug("query: %s", query)
    sql_lyr = data.ExecuteSQL(query)
    features = []
    if data is not None:
        LOGGER.info('query is finished, lets start the loop')
        for feature in tqdm(sql_lyr, desc=f'extract {geo_type}'):
            try:
                fields = [feature.GetField(key) for key in
                          ['osm_id', *constraint_dict['osm_keys']]]
                wkb = feature.geometry().ExportToWkb()
                geom = shapely.wkb.loads(bytes(wkb))
                if geom is None:
                    continue
                fields.append(geom)
                features.append(fields)
            except Exception as exc:
                LOGGER.info('%s - %s', exc.__class__, exc)
                LOGGER.warning("skipped OSM feature")
    else:
        LOGGER.error("""Nonetype error when requesting SQL. Check the
                     query and the OSM config file under the respective
                     geometry - perhaps key is unknown.""")

    return gpd.GeoDataFrame(
        features, columns=['osm_id', *constraint_dict['osm_keys'], 'geometry'],
        crs='epsg:4326')

# TODO: decide on name of wrapper, which categories included & what components fall under it.
def extract_cis(osm_path, ci_type):
    """
    A wrapper around extract() to conveniently extract map info for a
    selection of  critical infrastructure types from the given osm.pbf file.
    No need to search for osm key/value tags and relevant geometry types.
    Parameters
    ----------
    osm_path : str or Path
        location of osm.pbf file from which to parse
    ci_type : str
        one of DICT_CIS_OSM.keys(), i.e. 'education', 'healthcare',
        'water', 'telecom', 'road', 'rail', 'air', 'gas', 'oil', 'power',
        'wastewater', 'food'
    See also
    -------
    DICT_CIS_OSM for the keys and key/value tags queried for the respective
    CIs. Modify if desired.
    """
    # features consisting in points and multipolygon results:
    if ci_type in ['healthcare','education','food', 'buildings']:
        gdf = pd.concat([
            extract(osm_path, 'points', DICT_CIS_OSM[ci_type]['osm_keys'],
                    DICT_CIS_OSM[ci_type]['osm_query']),
            extract(osm_path, 'multipolygons', DICT_CIS_OSM[ci_type]['osm_keys'],
                    DICT_CIS_OSM[ci_type]['osm_query'])
            ])

    # features consisting in multipolygon results:
    elif ci_type in ['air']:
        gdf = extract(osm_path, 'multipolygons', 
                      DICT_CIS_OSM[ci_type]['osm_keys'],
                      DICT_CIS_OSM[ci_type]['osm_query'])

    # features consisting in points, multipolygons and lines:
    elif ci_type in ['gas','oil','telecom','water','wastewater','power',
                     'rail','road', 'main_road']:
        gdf =  pd.concat([
            extract(osm_path, 'points', DICT_CIS_OSM[ci_type]['osm_keys'],
                    DICT_CIS_OSM[ci_type]['osm_query']),
            extract(osm_path, 'multipolygons', DICT_CIS_OSM[ci_type]['osm_keys'],
                             DICT_CIS_OSM[ci_type]['osm_query']),
            extract(osm_path, 'lines', DICT_CIS_OSM[ci_type]['osm_keys'],
                             DICT_CIS_OSM[ci_type]['osm_query'])
            ])
    else:
        LOGGER.warning('feature not in DICT_CIS_OSM. Returning empty gdf')
        gdf = gpd.GeoDataFrame()
    return gdf



"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
test extraction functions
"""

import unittest
import geopandas as gpd
import numpy as np
from osm_flex.extract import extract, extract_cis, _query_builder
from pathlib import Path

PATH_TEST_DATA = Path(__file__).parent / 'data' 
OSM_FILE = PATH_TEST_DATA / 'test.osm.pbf'

class TestExtractionFunctions(unittest.TestCase):

    def test_extract(self):
        """
        test function extract()
        """
        gdf_mp = extract(OSM_FILE, 'multipolygons',  ['name', 'building'], 
                         "building='yes'")
        self.assertIsInstance(gdf_mp, gpd.GeoDataFrame)
        self.assertEqual(set(gdf_mp.columns), 
                         set(['osm_id', 'building', 'name', 'geometry']))
        self.assertEqual(np.unique(gdf_mp.geometry.type), 
                         np.array(['MultiPolygon']))
        self.assertTrue(len(gdf_mp)==4202)
        
        gdf_line = extract(OSM_FILE, 'lines', ['name', 'highway'], 
                           "highway='residential'")
        self.assertIsInstance(gdf_line, gpd.GeoDataFrame)
        self.assertEqual(set(gdf_line.columns), 
                         set(['osm_id', 'name', 'highway', 'geometry']))
        self.assertEqual(np.unique(gdf_line.geometry.type), 
                         np.array(['LineString']))
        self.assertTrue(len(gdf_line)==1807)
        
        # TODO: test with invalid geo_type

    def test_extract_cis(self):
        """
        test function extract_cis()
        """
        gdf_schools = extract_cis(OSM_FILE, 'education')
        self.assertTrue(np.any([entry =='college' for entry in gdf_schools.amenity]))
        self.assertTrue(np.any([entry =='school' for entry in gdf_schools.amenity]))
        self.assertTrue(len(gdf_schools)==215)
        self.assertTrue('amenity' in gdf_schools.columns)
        
        gdf_roads = extract_cis(OSM_FILE, 'road')
        self.assertTrue(np.any([entry =='tertiary' for entry in gdf_roads.highway]))
        self.assertTrue(np.any([entry =='residential' for entry in gdf_roads.highway]))
        self.assertTrue(len(gdf_roads)==2603)
        self.assertTrue('name' in gdf_roads.columns)
        
        # TODO: test with invalid ci-argument
        
    def test__query_builder(self):
        """
        test function _query_builder()
        """
        geo_type = 'points'
        constraint_dict = {'osm_keys': ['name', 'highway'],
                           'osm_query' :  "highway='residential'"}
        query = _query_builder(geo_type, constraint_dict)
        self.assertEqual(query,
                         "SELECT osm_id,name,highway FROM points WHERE highway='residential'")
      
    def test_qb_error(self):
        geo_type = 'points'
        constraint_dict2 = {'wrong_key': ['name', 'highway'],
                           'osm_query' :  "highway='residential'"}
        
        with self.assertRaises(KeyError) as context:
            _query_builder(geo_type, constraint_dict2)
        # TODO: how to check now that raises either keyerror or key error with message
        #self.assertTrue('osm_keys' in context.exception)
        #self.assertTrue(KeyError in context.exception)
       
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestExtractionFunctions)
    unittest.TextTestRunner(verbosity=2).run(TESTS)
